import 'dart:async';

import 'package:flutter/cupertino.dart';
import 'package:flutter/services.dart';
import 'package:flutter_webview_desktop/src/desktop_webview_enums.dart';
import 'package:flutter_webview_desktop/src/print.dart';
import 'package:rxdart/rxdart.dart';
import 'package:webview_flutter_platform_interface/webview_flutter_platform_interface.dart';

class DesktopWebviewValue {
  const DesktopWebviewValue({
    required this.isInitialized,
  });

  final bool isInitialized;

  DesktopWebviewValue copyWith({
    bool? isInitialized,
  }) {
    return DesktopWebviewValue(
      isInitialized: isInitialized ?? this.isInitialized,
    );
  }

  DesktopWebviewValue.uninitialized()
      : this(
          isInitialized: false,
        );
}

const String _pluginChannelPrefix = 'flutter_webview_desktop';
const MethodChannel _pluginChannel = MethodChannel(_pluginChannelPrefix);

class DesktopWebview extends ValueNotifier<DesktopWebviewValue> {
  DesktopWebview(this._javascriptChannelRegistry) : super(DesktopWebviewValue.uninitialized());

  int _textureId = 0;
  bool _isDisposed = false;
  late Completer<void> _creatingCompleter;
  final Completer<int> _texureIdCompleter = Completer();

  final JavascriptChannelRegistry _javascriptChannelRegistry;

  // 消息通道
  late MethodChannel _methodChannel;
  late EventChannel _eventChannel;

  Future<void> get ready => _creatingCompleter.future;
  Future<int> get textureId {
    return _texureIdCompleter.future;
  }

  Set<String> _extractChannelNames(Set<JavascriptChannel>? channels) {
    final Set<String> channelNames = channels == null ? <String>{} : channels.map((JavascriptChannel channel) => channel.name).toSet();
    return channelNames;
  }

  Future<void> updateJavascriptChannels(Set<JavascriptChannel>? newChannels) async {
    final Set<String> currentChannels = _javascriptChannelRegistry.channels.keys.toSet();
    final Set<String> newChannelNames = _extractChannelNames(newChannels);
    final Set<String> channelsToAdd = newChannelNames.difference(currentChannels);
    final Set<String> channelsToRemove = currentChannels.difference(newChannelNames);
    if (channelsToRemove.isNotEmpty) {
      await removeJavascriptChannels(channelsToRemove);
    }
    if (channelsToAdd.isNotEmpty) {
      await addJavascriptChannels(channelsToAdd);
    }
    _javascriptChannelRegistry.updateJavascriptChannelsFromSet(newChannels);
  }
  // //鼠标处理
  // final StreamController<SystemMouseCursor> _cursorStreamController = StreamController<SystemMouseCursor>.broadcast();
  //
  // /// A stream reflecting the current cursor style.
  // Stream<SystemMouseCursor> get _cursor => _cursorStreamController.stream;

  final BehaviorSubject<String> _urlStreamController = BehaviorSubject<String>();

  /// A stream reflecting the current URL.
  Stream<String> get url => _urlStreamController.stream;
  Future<String> get currentUrl => Future.value(_urlStreamController.value);

  final BehaviorSubject<DesktopWebviewLoadState> _loadStateStreamController = BehaviorSubject<DesktopWebviewLoadState>();

  /// A stream reflecting the current loading state.
  Stream<DesktopWebviewLoadState> get loadingState => _loadStateStreamController.stream;

  final BehaviorSubject<String> _titleStreamController = BehaviorSubject<String>();

  Stream<String> get title => _titleStreamController.stream;
  Future<String> get currentTitle => Future.value(_titleStreamController.value);

  final BehaviorSubject<DesktopWebviewClientHistoryChanged> _historyChangedStreamController = BehaviorSubject<DesktopWebviewClientHistoryChanged>();

  Stream<DesktopWebviewClientHistoryChanged> get historyChanged => _historyChangedStreamController.stream;

  Future<bool> get canGoBack => Future.value(_historyChangedStreamController.value.canGoBack);
  Future<bool> get canGoForward => Future.value(_historyChangedStreamController.value.canGoForward);

  Future<void> initialize(int width, int height, String initialUrl, double devicePixelRatio) async {
    if (_isDisposed) {
      return;
    }

    _loadStateStreamController.add(DesktopWebviewLoadState.start);
    final result = await _pluginChannel
        .invokeMapMethod<String, dynamic>('initialize', {'width': width, 'height': height, 'initialUrl': initialUrl, 'devicePixelRatio': devicePixelRatio});
    try {
      _creatingCompleter = Completer();
      if (result != null) {
        _textureId = result['textureId'];
        printLog('initialize textureId = $_textureId');
        _texureIdCompleter.complete(_textureId);
        value = value.copyWith(isInitialized: true);

        _methodChannel = MethodChannel('$_pluginChannelPrefix/$_textureId');
        _eventChannel = EventChannel('$_pluginChannelPrefix/$_textureId/events');

        ///event 消息的回调处理
        _eventChannel.receiveBroadcastStream().listen((event) {
          printLog('===== event $event');

          final map = event as Map<dynamic, dynamic>;
          switch (map['type']) {
            case 'urlChanged':
              _urlStreamController.add(map['value']);
              break;
            case 'onCursorChanged':
              // SystemMouseCursor systemMouseCursor = mouserCursors[map['value']] ?? SystemMouseCursors.basic;
              //
              // printLog('====== systemMouseCursor ${systemMouseCursor.toStringShort()}');
              // _cursorStreamController.add(systemMouseCursor);
              break;
            case 'loadStateChanged':
              final value = DesktopWebviewLoadState.values[map['value']];
              _loadStateStreamController.add(value);
              break;
            case 'titleChanged':
              _titleStreamController.add(map['value']);
              break;
            case 'onHistoryChanged':
              final value = DesktopWebviewClientHistoryChanged(map['value']['canGoBack'], map['value']['canGoForward']);
              _historyChangedStreamController.add(value);
              break;
          }
        });

        _methodChannel.setMethodCallHandler((call) async {
          switch (call.method) {
            case 'javascriptChannelMessage':
              final String channel = call.arguments['channel']! as String;
              final String message = call.arguments['message']! as String;
              _javascriptChannelRegistry.onJavascriptChannelMessage(channel, message);
              // return true;
              return true;
            default:
              return null;
          }
        });
      }
    } catch (e) {
      _creatingCompleter.completeError(e);
      return;
    }
    _creatingCompleter.complete();
    return _creatingCompleter.future;
  }

  @override
  Future<void> dispose() async {
    await _creatingCompleter.future;
    if (!_isDisposed) {
      printLog('dispose _pluginChannel $_textureId');
      await _pluginChannel.invokeMapMethod('dispose', _textureId);
      _isDisposed = true;
    }
    _isDisposed = true;

    super.dispose();
  }

  ///加载URL
  Future<void> loadUrl(String url, Map<String, String>? headers) async {
    if (_isDisposed) {
      return;
    }
    return _methodChannel.invokeMethod(
      'loadUrl',
      {'url': url, 'headers': headers},
    );
  }

  Future<void> loadRequest(Map<String, dynamic> webviewRequest) async {
    // TODO: implement loadRequest
    throw UnimplementedError();
    // if (_isDisposed) {
    //   return;
    // }
    // return _methodChannel.invokeMethod(
    //   'loadRequest',
    //   webviewRequest,
    // );
  }

  /// Loads a document from the given string.
  Future<void> loadHtmlString(String content, {String? baseUrl}) async {
    if (_isDisposed) {
      return;
    }
    return _methodChannel.invokeMethod('loadHtmlString', content);
  }

  /// Reloads the current document.
  Future<void> reload() async {
    if (_isDisposed) {
      return;
    }
    return _methodChannel.invokeMethod('reload');
  }

  /// Stops all navigations and pending resource fetches.
  Future<void> stop() async {
    if (_isDisposed) {
      return;
    }
    return _methodChannel.invokeMethod('stop');
  }

  /// Navigates the WebView to the previous page in the navigation history.
  Future<void> goBack() async {
    if (_isDisposed) {
      return;
    }
    return _methodChannel.invokeMethod('goBack');
  }

  /// Navigates the WebView to the next page in the navigation history.
  Future<void> goForward() async {
    if (_isDisposed) {
      return;
    }
    return _methodChannel.invokeMethod('goForward');
  }

  Future<void> scrollDelta(Offset orignal, Offset delta) async {
    if (_isDisposed) {
      return;
    }
    printLog('scrollDelta,orignal = ${orignal.toString()} delta = ${delta.toString()}');

    return _methodChannel
        .invokeMethod("scrollDelta", {"x": orignal.dx.toInt(), "y": orignal.dy.toInt(), "deltaX": delta.dx.toInt(), "deltaY": delta.dy.toInt()});
  }

  Future<void> clearCache() async {
// TODO: implement clearCache
    throw UnimplementedError();
  }

  Future<void> addJavascriptChannels(Set<String> javascriptChannelNames) async {
    if (_isDisposed) {
      return;
    }

    if (javascriptChannelNames.isEmpty) {
      return;
    }
    printLog('addJavascriptChannels= $javascriptChannelNames');
    return _pluginChannel.invokeMethod('addJavascriptChannels', javascriptChannelNames.toList().first);
  }

  Future<void> removeJavascriptChannels(Set<String> javascriptChannelNames) async {
    if (_isDisposed) {
      return;
    }

    if (javascriptChannelNames.isEmpty) {
      return;
    }
    printLog('removeJavascriptChannels= $javascriptChannelNames');
    return _methodChannel.invokeMethod('removeJavascriptChannels', javascriptChannelNames.toList().first);
  }

  Future<String> evaluateJavascript(String javascript) async {
    if (_isDisposed) {
      return '';
    }
    if (javascript.isEmpty) {
      return '';
    }

    await _methodChannel.invokeMethod('evaluateJavascript', <String, dynamic>{'url': _urlStreamController.value, 'code': javascript});
    return '';
  }

  Future<int> getScrollX() {
    // TODO: implement getScrollX
    throw UnimplementedError();
  }

  Future<int> getScrollY() {
    // TODO: implement getScrollY
    throw UnimplementedError();
  }

  Future<void> setWindowOpenPolicy(DesktopWebviewClientWindowOpenPolicy webviewClientWindowOpenPolicy) async {
    if (_isDisposed) {
      return;
    }
    printLog('_setWindowOpenPolicy,webviewClientWindowOpenPolicy = $webviewClientWindowOpenPolicy');

    return _methodChannel.invokeMethod('setWindowOpenPolicy', webviewClientWindowOpenPolicy.index);
  }

  /// Indicates whether the specified [button] is currently down.
  Future<void> setMouseClick(DesktopMousePointerButton button, bool isDown, Offset position) async {
    if (_isDisposed) {
      return;
    }

    printLog('===_setMouseClick button = $button isDown = $isDown x = ${position.dx.toInt()}  y = ${position.dy.toInt()}');
    if (position.dx < 0 || position.dy < 0) {
      return;
    }
    return _methodChannel
        .invokeMethod('setMouseClick', <String, dynamic>{'button': button.index, 'isDown': isDown, 'x': position.dx.toInt(), 'y': position.dy.toInt()});
  }

  Future<void> setViewSize(Size size, double devicePixelRatio) async {
    if (_isDisposed) {
      return;
    }
    printLog('===setSize  size ${size.toString()} devicePixelRatio = $devicePixelRatio');
    return _methodChannel.invokeMethod('setViewSize', {'width': size.width.toInt(), 'height': size.height.toInt(), 'devicePixelRatio': devicePixelRatio});
  }
}
